#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
extern "C" int abs(...);
#endif

#include "compilerApi.h"
#include "strategy.h"
#include "math.h"

#include "anls_att.h"
#include "detect_comm.h"
#include "indep.h"

/*******************************************************************************
		function to deal with remap communication
*******************************************************************************/

// if RHS have different type as LHS , reset type as RHS
void reset_indepdo_tdd_type(SgStatement *s,int counter)
{
  Indepdo_reference * lhsref=(Indepdo_reference *)s->attributeValue(0, Indepdo_Lref);
  SgArrayRefExp * lhsarray=(SgArrayRefExp *)lhsref->getexpression();
  if(lhsarray->numberOfAttributes(KEPT_ARRAY_REF)) lhsarray = (SgArrayRefExp *) lhsref->getexpression()->attributeValue(0,KEPT_ARRAY_REF);
  SgSymbol * lhsarraysymb=lhsarray->symbol(); 

  Indepdo_reference * rhsref=(Indepdo_reference *)s->attributeValue(counter, Indepdo_Ref);
  SgArrayRefExp * rhsarray = (SgArrayRefExp *) rhsref->getexpression();
  SgSymbol * rhsarraysymb=rhsref->getexpression()->symbol(); 

  SgSymbol * tdd = s->getSymbolInCurrentScope(mixstring, "tdd", NULL,counter);

  SgType *ltp = lhsarraysymb->type();
  SgType *rtp = rhsarraysymb->type();
  if (ltp->hasBaseType()) ltp = ltp->baseType();
  if (ltp->hasBaseType()) ltp = ltp->baseType();//in case of T_KIND
  if (rtp->hasBaseType()) rtp = rtp->baseType();
  if (rtp->hasBaseType()) rtp = rtp->baseType();
  if ( ltp->variant() != rtp->variant() ) {
	SgStatement *stmt = make_set_type(s, tdd, rtp);
    	s->insertStmtBefore(*stmt, *s->controlParent());
  }
}

// process: broadcast to dimension
// create copy of the range from the LHS
// and set outofcore if needed,this function is not used currently 
void add_indepdo_range_copy(SgStatement *s,int counter)
{
  int i;

  Indepdo_header_def * ihdef = (Indepdo_header_def *) s->attributeValue(0, IHEDR_Def);

  Indepdo_reference * iref= (Indepdo_reference *) s->attributeValue(counter, Indepdo_Ref);
  Indepdo_reference * lhsref= (Indepdo_reference *) s->attributeValue(0, Indepdo_Lref);

  SgArrayRefExp * array = (SgArrayRefExp *) iref->getexpression();
  SgSymbol * arraysymb  = iref->getexpression()->symbol();

  SgArrayRefExp * lhsarray;
  if(array->numberOfAttributes(KEPT_ARRAY_REF)) lhsarray = (SgArrayRefExp *) lhsref->getexpression()->attributeValue(KEPT_ARRAY_REF);
  else lhsarray=(SgArrayRefExp *) lhsref->getexpression();
  SgSymbol * lhsarraysymb  = lhsarray->symbol();

    // storage must be same as LHS  ???
  int storage = getArrayStorage(arraysymb);

  SgSymbol * dad_parent = s->getSymbolInCurrentScope(mixstring, "sdd", NULL,0);

  int dimcounter;
  if (storage == COLUMN_MAJOR) 
  {
    dimcounter=0;
    for(i=0; i<MAX_ALIGN_DIM;i++)
    {
	// dim is the correspondent dimension of lhs
        int dim=((forall_reference *)iref)->getAmount(i);
	if (dim==-1) continue;
	dimcounter++;

	// pcrc_set_array_copy
        SgSymbol * dad=s->getSymbolInCurrentScope(mixstring, "tdd", NULL,counter);
	SgStatement *array_copy = make_set_array_copy(s,dad,dimcounter,dad_parent,dim+1);

        s->insertStmtBefore(*array_copy, *s->controlParent());
    }
  }
  else
  {
    dimcounter = 1;
    for (i=0; i< MAX_ALIGN_DIM; i++) 
	if (((forall_reference *)iref)->getAmount(i) != -1) dimcounter++;
    for(i=MAX_ALIGN_DIM; i>=0; i--)
    {
	// dim is the correspondent dimension of lhs
        int dim=((forall_reference *)iref)->getAmount(i);
	if (dim==-1) {
	  continue;
	}
	dimcounter--;

	// pcrc_set_array_copy
        SgSymbol * dad=s->getSymbolInCurrentScope(mixstring, "tdd", NULL,counter);
	SgStatement *array_copy = make_set_array_copy(s,dad,dimcounter,dad_parent,dim+1);

        s->insertStmtBefore(*array_copy, *s->controlParent());
    }
  }

    // as if lsh is in_core
  if (array->numberOfAttributes(SET_INCORE) > 0);
  else
  if (is_out_of_core(lhsarraysymb))
  {
        SgSymbol * dad=s->getSymbolInCurrentScope(mixstring, "tdd", NULL,counter);
        SgStatement * ts=make_oc_init(s,dad,IN_CORE_SIZE);
        s->insertStmtBefore(*ts, *s->controlParent());
  }

}

// create tdd(target dad) for remap
void create_remap_target (SgStatement *s,int counter)
{
  int i;

  Indepdo_header_def * ihdef = (Indepdo_header_def *) s->attributeValue(0, IHEDR_Def);

  Indepdo_reference * lhsref= (Indepdo_reference *)s->attributeValue(0, Indepdo_Lref);
 
  SgArrayRefExp * lhsarray = (SgArrayRefExp *) lhsref->getexpression();
  stack stk;
  if(lhsarray->numberOfAttributes(KEPT_ARRAY_REF)) {//LHS already changed 
  	SgArrayRefExp *old_lhsarray=(SgArrayRefExp *) lhsref->getexpression()->attributeValue(0,KEPT_ARRAY_REF);
	stk.push(lhsarray->copyPtr());
	lhsarray->replace(*old_lhsarray);//change to old
  }
  SgSymbol * lhsarraysymb  = lhsarray->symbol();

  Indepdo_reference * rhsref = (Indepdo_reference *)s->attributeValue(counter, Indepdo_Ref);
  SgArrayRefExp * rhsarray = (SgArrayRefExp *) rhsref->getexpression();
  SgSymbol * rhsarraysymb  = rhsarray->symbol();

  int storage = getArrayStorage(rhsarraysymb);

  SgArrayRefExp * array = (SgArrayRefExp *) lhsref->getexpression();
  SgSymbol * arraysymb  = lhsref->getexpression()->symbol();

  SgSymbol * tddsymb=s->getSymbolInCurrentScope(mixstring, "tdd", NULL,counter);

    // pcrc_new_array_section
  {
    int dim = ihdef->getindexnumber();
    SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
    SgStatement* assign = make_new_array_section(s, tddsymb, dim , dad);
    s->insertStmtBefore(*assign, *s->controlParent());
  }

    // set tdd as IN_CORE when possible
  {
    if (is_out_of_core(lhsarraysymb) && !is_out_of_core(rhsarraysymb) )
    {
	  SgStatement * stmt = make_set_ic(s,tddsymb);
    	  s->insertStmtBefore(*stmt, *s->controlParent());
    }
  }

  int dimcounter = 0;

    // set_array_triplet or set_array_scalar
  for (i=0; i<array->numberOfSubscripts(); i++) {
    int index = lhsref->indexNumberOnDim(i);

    if (index!=-1) {
        // set_array_triplet
      dimcounter++;

      // the following subprogram add start, end, stride;
      SgExpression *start, *end, *stride;
      {
	start = lhsref->getoffset(i,index);
	start = start+((*lhsref->getstride(i,index))*(*ihdef->getindextriple(index,0)));
	analyse_expression(s, start);

	end = lhsref->getoffset(i,index);
	end = end+((*lhsref->getstride(i,index))*(*ihdef->getindextriple(index,1)));
	analyse_expression(s, end);

	if (ihdef->getindextriple(index, 2)!=NULL) 
	  stride = &(*lhsref->getstride(i,index))*(*ihdef->getindextriple(index,2));
	else 
	  stride = lhsref->getstride(i,index);
	analyse_expression(s, stride);
      }

      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());

      SgStatement *funcall = make_set_array_triplet(s,tddsymb,dimcounter,start,
							end,stride,dad,i+1);
      s->insertStmtBefore(*funcall, *s->controlParent());
    }
    else {
	// set_array_scalar
      SgExpression * scalar = ( (SgArrayRefExp *) lhsref->getexpression() ) ->subscript(i);
      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
      SgStatement *funcall = make_set_array_scalar(s,tddsymb,scalar,dad,i+1);
      s->insertStmtBefore(*funcall, *s->controlParent());
    }
  }

    // allocate memory according to RHS symbol's type, even if create tdd as LHS
  SgSymbol * mHandle;
  {
    SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", NULL, counter);

#ifdef MEM_ALLOC
  	mHandle = s->getSymbolInCurrentScope(mixstring, "mh", NULL, counter);
	SgStatement *alloc=make_stack_alloc(s,rhsarraysymb->type(),base,tddsymb,mHandle);
#else
	SgStatement *alloc=make_alloc(s,rhsarraysymb->type(),base,tddsymb);
#endif

    s->insertStmtBefore(*alloc, *s->controlParent());
  }

    // set type as RHS
  {
    reset_indepdo_tdd_type(s,counter);
  }

    // reset base;
  {
    SgSymbol * basesym = s->getSymbolInCurrentScope(mixstring, "tbs", NULL,counter);

    SgStatement *funcall=make_reset_array_base(s,tddsymb,rhsarraysymb->type(),basesym, mHandle);

    s->insertStmtBefore(*funcall, *s->controlParent());
  }

    // call delete_array;
  {
    SgStatement *delete_stmt = make_delete_array(s,tddsymb);
    s->addAttribute(INSERTED_AFTER_STMT,delete_stmt, sizeof(SgStatement));
  }

    // call real free;
  {
#ifdef MEM_ALLOC
    SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "mh", NULL, counter);
#else
    SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", NULL, counter);
#endif
    SgStatement* free_stmt = make_free(s,rhsarraysymb->type(),base);
    s->addAttribute(INSERTED_AFTER_STMT,free_stmt, sizeof(SgStatement));
  }
  if(lhsarray->numberOfAttributes(KEPT_ARRAY_REF)){//replace back
	lhsarray->replace(*stk.top_exp());
  }
}

// create pcrc_remap
void add_indepdo_remap(SgStatement * s, int counter)
{

// decide if we must transpose the remap

  int remap=0;
  int dd[MAX_ALIGN_DIM+1];
  int dno=0;
  int i;
  Indepdo_reference * lhsref = (Indepdo_reference *) s->attributeValue(0, Indepdo_Lref);
  Indepdo_reference * rhsref = (Indepdo_reference *) s->attributeValue(counter, Indepdo_Ref);
  for (i=0; i<MAX_ALIGN_DIM ; i++)
  {
	dd[i]=-1;
	if (((forall_reference *)rhsref)->getAmount(i)!=-1)
	{
		dno++;
		int nodim=0;
		int amount=((forall_reference *)rhsref)->getAmount(i);
		// delete the no scalar dim
		for (int k=0; k<amount; k++)
			if (lhsref->indexNumberOnDim(k)==-1) nodim++;
		dd[dno-1]=amount-nodim;
	}
  }
  dd[dno]=-1;

  i=0;
  while (i<dno-1 && dd[i]<dd[i+1]) i++;
  if (i==dno-1 || dno==0) remap=1;
    // if BROADCAST_DIM , set_array_align insure the mapping.
  if (rhsref->getCommunication()==BROADCAST_DIM) remap=1;

//decide complete

  SgSymbol * symb=s->getSymbolInCurrentScope(mixstring, remap?"remap":"remap_transpose", NULL);
  SgCallStmt * funcall_read= new SgCallStmt(*symb);
  SgCallStmt * funcall_write= new SgCallStmt(*symb);

  SgSymbol *dadsec = s->getSymbolInCurrentScope(mixstring, "tdd", NULL, counter);
  SgExpression * dadexp = new SgVarRefExp(dadsec);

  SgSymbol *dadsrc = s->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);
  SgExpression * dadsecexp = new SgVarRefExp(dadsrc);

  funcall_read->addArg(*dadexp);
  funcall_read->addArg(*dadsecexp);

  funcall_write->addArg(*dadsecexp);
  funcall_write->addArg(*dadexp);

  delete dadexp;
  delete dadsecexp;

  if (remap==0)
    for (i=0;i<=dno; i++)
    {
	SgExpression *dexp = new SgValueExp(dd[i]);
  	funcall_read->addArg(*dexp);
	delete dexp;
    }

  if((rhsref->get_ref_attrib()|rhsref->added_ref_attrib)&Right_Bit||rhsref->getexpression()->numberOfAttributes(CONDITION_ASSIGN))
	  s->insertStmtBefore(*funcall_read, *s->controlParent());
  if((rhsref->get_ref_attrib()|rhsref->added_ref_attrib)&Left_Bit)
	  s->addAttribute(INSERTED_AFTER_STMT,funcall_write,sizeof(SgCallStmt *));
}

/*******************************************************************************
		function to deal with shift communication
			do_write_halo(s)
*******************************************************************************/

void get_indepdo_shift_array(SgStatement *s)
{
  for (int i=0; i<s->numberOfAttributes(Indepdo_Ref); i++) {
    Indepdo_reference * iref = (Indepdo_reference *)s->attributeValue(i, Indepdo_Ref);

    if (iref->getCommunication()==SHIFT ){
      int found =0;

      for ( int j=0; j<s->numberOfAttributes(SHIFTARRAY_Def); j++) {
	SgSymbol *array = (SgSymbol *)s->attributeValue(j, SHIFTARRAY_Def);
	
	if (array==iref->getexpression()->symbol()) {
	  found =1;
	  break;
	};
      };

      if (!found) {
	s->addAttribute(SHIFTARRAY_Def, iref->getexpression()->symbol(), sizeof(SgSymbol)); 
      }
      
    }
  }
}

int gcd(int a,int b){
	ASSERT(a>0&&b>0);
	int k=a;
	if(k>b) k=b;
	while(k>1){
		if((a/k)*k==a&&(b/k)*k==b)
			return k;
		k--;
	}
	return 1;
}

int need_rw(Shift_def *sdef,int rw_f){
	if(rw_f==0){//read
		return (sdef->shift_iref->get_ref_attrib()&Right_Bit)||sdef->shift_iref->getexpression()->numberOfAttributes(CONDITION_ASSIGN);
	}
	else if(rw_f==1) return sdef->shift_iref->get_ref_attrib()&Left_Bit;
	else ASSERT(0);
}

// create a Shadow_def(LNEEDED) for array
void decide_indepdo_shadow(SgStatement * s)
{
  for (int k=0; k<s->numberOfAttributes(SHIFTARRAY_Def); k++) {
    SgSymbol *array = (SgSymbol *)s->attributeValue(k, SHIFTARRAY_Def);

  for(int rw_f=0;rw_f<2;rw_f++){//create read and write shift
      Shift_def * dsdef = new Shift_def (LNEEDED);
	dsdef->rw_flag=rw_f;
      int found=0;
      for (int numOfDim=0; numOfDim < MAX_ALIGN_DIM; numOfDim++) {
	int lsize=0;
	int usize=0;
	
	for ( int numS=0; numS<array->numberOfAttributes(Shift_Def); numS++) {
	  Shift_def * sdef = (Shift_def *)array->attributeValue(numS, Shift_Def);
	  if (sdef->getTag()==DETECTED && sdef->getScope()==s&&need_rw(sdef,rw_f)) {
	    found=1;
	    int tmp;
	    if ((tmp=sdef->getAmt(numOfDim)) < 0){
			if(-tmp>lsize) lsize = -tmp;
	    }
	    else if (tmp > 0){
	    		if(tmp>usize) usize = tmp;
	    }
	}
        }
	dsdef->setLower(numOfDim, lsize);
	dsdef->setUpper(numOfDim, usize);      
    }
        if(!found) continue; 
	int start=0;
	Shift_def *sdef=(Shift_def *)array->attributeValue(0,Shift_Def);
  	while(!(sdef->getTag()==DETECTED && sdef->getScope()==s)&&need_rw(sdef,rw_f)){
		if(start==array->numberOfAttributes(Shift_Def)-1){
			sdef=NULL;
			break;
		}
		start++;
		sdef=(Shift_def *)array->attributeValue(start,Shift_Def);
	}
	ASSERT(sdef);
	for(int i=0;i<((SgArrayType *)array->type())->dimension();i++){
		SgExpression * l_exp=sdef->getDimRange(i)->lhs()->lhs();
		SgExpression * u_exp=sdef->getDimRange(i)->lhs()->rhs();
		if(!l_exp->isInteger()||!u_exp->isInteger()){
			SgExpression *range=new SgExpression(RANGE_OP,l_exp,u_exp,NULL);
			SgExpression *ddot=new SgExpression(DDOT,range,new SgValueExp(1),NULL);
			dsdef->setDimRange(i,ddot);
			continue;
		}
		int l=l_exp->valueInteger();
		int u=u_exp->valueInteger();
		int st=sdef->getDimRange(i)->rhs()?sdef->getDimRange(i)->rhs()->valueInteger():1;
		int got=0;
		for(int j=start+1;j<array->numberOfAttributes(Shift_Def);j++){
		        Shift_def *sdef=(Shift_def *)array->attributeValue(j,Shift_Def);
			if(!(sdef->getTag()==DETECTED && sdef->getScope()==s&&need_rw(sdef,rw_f) )) continue;
			SgExpression * l_exp=sdef->getDimRange(i)->lhs()->lhs();
			SgExpression * u_exp=sdef->getDimRange(i)->lhs()->rhs();
			if(!l_exp->isInteger()||!u_exp->isInteger()){
				SgExpression *range=new SgExpression(RANGE_OP,l_exp,u_exp,NULL);
				SgExpression *ddot=new SgExpression(DDOT,range,new SgValueExp(1),NULL);
				dsdef->setDimRange(i,ddot);
				got=1;
				break;
			}
			int tmp,tmp1;
			if((tmp=sdef->getDimRange(i)->lhs()->rhs()->valueInteger())>u)  u=tmp;
			tmp=sdef->getDimRange(i)->rhs()?sdef->getDimRange(i)->rhs()->valueInteger():1;
			tmp1=sdef->getDimRange(i)->lhs()->lhs()->valueInteger();
			if(st==1||tmp==1) st=1;
			else st=(tmp1-l)?gcd(gcd(st,tmp),abs(tmp1-l)):gcd(st,tmp);
			if(tmp1<l) l=tmp1;
		}
		if(got) continue;
		SgExpression *range=new SgExpression(RANGE_OP,new SgValueExp(l),new SgValueExp(u),NULL);
		SgExpression *ddot=new SgExpression(DDOT,range,new SgValueExp(st),NULL);
		dsdef->setDimRange(i,ddot);
	}
 
      //dsdef->show();
      dsdef->setScope(s);
      
      array->addAttribute(Shift_Def,(void *) dsdef, sizeof(Shift_def));
//      array->addAttribute(Shad_Def,(void *) dsdef, sizeof(Shift_def));
  }//read and write shift created  
} // end of for
}

void check_indepdo_ghost(SgStatement * s)
{

  get_indepdo_shift_array(s);

  decide_indepdo_shadow(s);
}


void create_shift_array (SgStatement *s, SgSymbol * arraysymb)
{
  int i;
  SgSymbol * tddsymb=s->getSymbolInCurrentScope(mixstring, "tdd", NULL);
  SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());

  for(int rw_f=0;rw_f<2;rw_f++){
  Shift_def *sdef=NULL;
  for(i=0;i<arraysymb->numberOfAttributes(Shift_Def);i++){
	Shift_def *shdef=(Shift_def *)arraysymb->attributeValue(i,Shift_Def);
	if(shdef->getTag()==LNEEDED&&shdef->getScope()==s&&shdef->rw_flag==rw_f){
		sdef=shdef;
		break;
	}
  }
  if(!sdef) continue;
  int dimcounter = 0;
  int dim=((SgArrayType *)arraysymb->type())->dimension();

  for (i=0; i<dim; i++) {
	SgExpression * l=sdef->getDimRange(i)->lhs()->lhs();
	SgExpression * u=sdef->getDimRange(i)->lhs()->rhs();
    if (*l==*u) ;
	else dimcounter++;
  }

  // pcrc_new_array_section()
  {
    SgStatement* assign = make_new_array_section(s, tddsymb, dimcounter , dad);
    if(rw_f==0) s->insertStmtBefore(*assign, *s->controlParent());
    else s->addAttribute(FUTURE_INSERT_AFTER,assign,sizeof(assign));
  }

  dimcounter = 0;

  for (i=0; i<dim; i++) {
	SgExpression * l=sdef->getDimRange(i)->lhs()->lhs();
	SgExpression * u=sdef->getDimRange(i)->lhs()->rhs();

    if (*l!=*u) {
	// pcrc_set_array_triplet()
      dimcounter++;

      SgExpression *start = sdef->getDimRange(i)->lhs()->lhs();
      analyse_expression(s, start);
      SgExpression *end  = sdef->getDimRange(i)->lhs()->rhs();
      analyse_expression(s, end);
      SgExpression *stride = sdef->getDimRange(i)->rhs();

      SgStatement *funcall = make_set_array_triplet(s, tddsymb, dimcounter,
				start,end,stride,dad,i+1);
      if(rw_f==0) s->insertStmtBefore(*funcall, *s->controlParent());
      else s->addAttribute(FUTURE_INSERT_AFTER,funcall,sizeof(funcall));
    }
    else {
	// pcrc_set_array_scalar()
      SgExpression * scalar = l;
      SgStatement *funcall = make_set_array_scalar(s,tddsymb,scalar,dad,i+1);
      if(rw_f==0) s->insertStmtBefore(*funcall, *s->controlParent());
      else s->addAttribute(FUTURE_INSERT_AFTER,funcall,sizeof(funcall));
      analyse_expression(s, scalar);
    }
  } // end of for
  }
}

void add_shift_write_halo(SgStatement *s, int counter)
{
      SgSymbol *array =(SgSymbol *)s->attributeValue(counter,SHIFTARRAY_Def);

	int read=0;
	int write=0;
      for(int i=0;i<array->numberOfAttributes(Shift_Def);i++){
      	Shift_def *sdef=(Shift_def *)array->attributeValue(i,Shift_Def);
	if(sdef->getTag()==LNEEDED && sdef->getScope()==s){
		if(sdef->rw_flag==0) read=1;
		if(sdef->rw_flag==1) write=1;	
	}
      }
      SgSymbol * dad=s->getSymbolInCurrentScope(mixstring, "tdd", NULL);  
      SgSymbol *gtl=s->getSymbolInCurrentScope(mixstring, "gtl", array->identifier());
      SgSymbol *gtu=s->getSymbolInCurrentScope(mixstring, "gtu", array->identifier());
	if(read){
		SgStatement *func = make_write_halo(s, dad,gtl, gtu);
		s->insertStmtBefore(*func, *s->controlParent());
		SgStatement *funcall = make_delete_array(s,dad);
		s->insertStmtBefore(*funcall, *s->controlParent());
	}
	if(write){
		SgStatement *func = make_write_back_halo(s, dad,gtl, gtu);
		SgStatement *funcall = make_delete_array(s,dad);
		s->addAttribute(INSERTED_AFTER_STMT,funcall,sizeof(funcall));
		s->addAttribute(INSERTED_AFTER_STMT,func,sizeof(func));
		for(int i=s->numberOfAttributes(FUTURE_INSERT_AFTER)-1;i>-1;i--){
			SgStatement * stmt=(SgStatement *)s->attributeValue(i,FUTURE_INSERT_AFTER);
			s->addAttribute(INSERTED_AFTER_STMT,stmt,sizeof(stmt));
  		}
		s->deleteAllAttribute(FUTURE_INSERT_AFTER);
	}
}

void define_indepdo_shift_amount(SgStatement * s, int counter)
{
  SgSymbol *array = (SgSymbol *)s->attributeValue(counter, SHIFTARRAY_Def);

  for(int rw_f=0;rw_f<2;rw_f++){
  for (int i=0; i<array->numberOfAttributes(Shift_Def); i++) {
    Shift_def * sdef = (Shift_def *)array->attributeValue(i, Shift_Def);
    if (sdef->getTag()==LNEEDED && sdef->getScope()==s&&sdef->rw_flag==rw_f) {

      SgArrayType * arrayType=(SgArrayType *)array->type();
      int dim = arrayType->dimension();
	int dimcounter=0;
      for(int j=0; j<dim; j++){
	int l=sdef->getDimRange(j)->lhs()->lhs()->valueInteger();
	int u=sdef->getDimRange(j)->lhs()->rhs()->valueInteger();
	if(l==u) continue;

	dimcounter++;

	SgValueExp * lowerExp= new SgValueExp(sdef->getLower(j));
	SgValueExp * upperExp= new SgValueExp(sdef->getUpper(j));
	SgValueExp * dimExp=new SgValueExp(dimcounter);


	SgSymbol * gtl=s->getSymbolInCurrentScope(mixstring, "gtl", array->identifier());
	SgSymbol * gtu=s->getSymbolInCurrentScope(mixstring, "gtu", array->identifier());
      
	SgExpression * range = new SgExpression(RANGE_OP, NULL, dimExp, NULL);
	SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
	SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

	SgExpression * gtlarray = new SgExpression(ARRAY_REF, index, NULL, gtl);
	SgExpression * gtuarray = new SgExpression(ARRAY_REF, index, NULL, gtu);

	SgAssignStmt * gtlAssign = new SgAssignStmt(*gtlarray, *lowerExp);
	SgAssignStmt * gtuAssign = new SgAssignStmt(*gtuarray, *upperExp);

	if(rw_f==0) s->insertStmtBefore(*gtlAssign, *s->controlParent());
	else s->addAttribute(FUTURE_INSERT_AFTER,gtlAssign,sizeof(gtlAssign));
	if(rw_f==0) s->insertStmtBefore(*gtuAssign, *s->controlParent());
	else s->addAttribute(FUTURE_INSERT_AFTER,gtuAssign,sizeof(gtuAssign));

      }
      break;
    }
  }
  }
}

void do_write_halo(SgStatement *s)
{
  for (int i=0; i<s->numberOfAttributes(SHIFTARRAY_Def); i++) {
    SgSymbol *array = (SgSymbol *)s->attributeValue(i, SHIFTARRAY_Def);

    if (array->numberOfAttributes(Shift_Def)) {
	create_shift_array(s,array);
	define_indepdo_shift_amount(s, i);
	add_shift_write_halo(s, i);

    }
  }
}

void fill_shift_section(SgStatement *s,int counter,Shift_def *sdef,int num){
	Indepdo_header_def *ihdef=(Indepdo_header_def *)s->attributeValue(0,IHEDR_Def);
	Indepdo_reference *iref=(Indepdo_reference *)s->attributeValue(counter,Indepdo_Ref);
	SgArrayRefExp *array=(SgArrayRefExp *)iref->getexpression();
	SgSymbol *arraysym=array->symbol();
	for(int i=0;i<array->numberOfSubscripts();i++){
		SgExpression *lb,*ub,*s=NULL;
		int amt;
		int index;
		if((index=iref->indexNumberOnDim(i))==-1){
			amt=iref->getAmount(num,i);
			SgExpression *sub=array->subscript(i);
			if(sub->variant()==DDOT){
				if(sub->lhs()->lhs()){
					lb=sub->lhs()->lhs();
				}
				else{
					lb=get_lbound(arraysym,i);
				}
				if(sub->lhs()->rhs()){
					ub=sub->lhs()->rhs();
				}
				else{
					ub=get_ubound(arraysym,i);
				}
				if(sub->rhs())
					s=sub->rhs();
				else s=new SgValueExp(1);
			}
			else{
				lb=array->subscript(i)->copyPtr();
				ub=lb->copyPtr();
			}
			if(sub->variant()==DDOT){
				if(amt==0);
				else
				if(amt>0){
					int tmp,step=s->valueInteger();
					
					if(amt/step*step!=amt) tmp=amt/step*step+step;
					else tmp=amt;

					lb=&(*ub-*new SgValueExp(tmp));
					lb=calculateExp(lb);
				}
				else{
					int tmp,step=s->valueInteger();
					
					if((-amt)/step*step!=(-amt)) tmp=(-amt)/step*step+step;
					else tmp=-amt;
                                        ub=&(*lb+*new SgValueExp(tmp));
                                        ub=calculateExp(ub);
				}
			}
			else{
				if(amt==0) ;
				else
				if(amt>0){
					lb=&(*ub-*new SgValueExp(amt));
					lb=calculateExp(lb);
					s=new SgValueExp(amt);
				}
				else{
                                        ub=&(*lb+*new SgValueExp(-amt));
                                        ub=calculateExp(ub);
                                        s=new SgValueExp(-amt);
				}
			}
		}
		else{
			amt=iref->getAmount(num,i);
			SgExpression *index_lbExp=ihdef->getindextriple(index,0)->copyPtr();
			SgExpression *index_ubExp=ihdef->getindextriple(index,1)->copyPtr();
			if(!index_lbExp->isInteger()||!index_ubExp->isInteger()||array->subscript(i)->numberOfAttributes(SUB_RANGE)){
				SgExpression *range=new SgExpression(RANGE_OP,get_lbound(arraysym,i),get_ubound(arraysym,i),NULL);
				SgExpression *ddot=new SgExpression(DDOT,range,NULL,NULL);
				sdef->setDimRange(i,ddot);
				sdef->setAmt(i,amt);
				continue;
			}
			SgExpression *index_strideExp=ihdef->getindextriple(index,2);
			if(index_strideExp) ASSERT(index_strideExp->isInteger());
			int index_stride;
			if(index_strideExp) index_stride=index_strideExp->valueInteger(); 
			else index_stride=1;
			int index_ub=index_ubExp->valueInteger();
			int index_lb=index_lbExp->valueInteger();
			SgExpression *strideExp=iref->getstride(i,index);
			if(!strideExp->isInteger()) ASSERT(0);
			int stride=strideExp->valueInteger();
			if(stride<0){
				index_stride=-index_stride;
				index_lb=-index_lb;
				index_ub=-index_ub;
				stride=-stride;
			}
			if(index_stride<0){
				int tmp=index_lb-index_ub;
				tmp=tmp/(-index_stride)*index_stride+index_lb;
				index_ub=index_lb;
				index_lb=tmp;
				index_stride=-index_stride;
			}
			ASSERT(index_stride>0&&stride>0);
			index_ub=(index_ub-index_lb)/index_stride*index_stride+index_lb;
/*			if(amount>0){
				int tmp=amount/(stride*index_stride);
				if(tmp*stride*index_stride<amount){
					amt=tmp+1;
				}
				else amt=tmp;
			}
			else if(amount==0) amt=0;
			else{
				int tmp=(-amount)/(stride*index_stride);
				if(tmp*stride*index_stride<-amount){
					amt=-(tmp+1);
				}
				else amt=-tmp;
			}*/
			SgExpression *offsetExp=iref->getoffset(i,index);
			int offset;
			if(offsetExp){
                        	if(!offsetExp->isInteger()) ASSERT(0);
				offset=offsetExp->valueInteger();
			}
			else offset=0;
			int l=offset+index_lb*stride;
//			printf("l:%d,offset:%d,index_lb:%d,stride:%d\n",l,offset,index_lb,stride);
			l=amt>0?l-amt:l;
			int u=index_ub*stride+offset;
			u=amt<0?u+(-amt):u;
//			printf("l:%d,u:%d,amt:%d\n",l,u,amt);
			lb=new SgValueExp(l);
			ub=new SgValueExp(u);
			s=new SgValueExp(stride*index_stride);
		}
		SgExpression *range=new SgExpression(RANGE_OP,lb,ub,NULL);
		SgExpression *ddot=new SgExpression(DDOT,range,s,NULL);
		sdef->setDimRange(i,ddot);
		sdef->setAmt(i,amt);
	}
}

void calc_shift_amount(detect_DIM *dad_x, detect_DIM *dad_y,int p,int lx,int a0, 
		int b0, int a1, int b1, int s, int tx, int ty, int *amount)
{
   int int_devide(int s1,int s2);
   int kx,ky_f,ky,kx_each,ky_each;

  switch(dad_x->distCode) {
		case 1 : 		// BLOCK Distribution
		    kx = dad_x->t_stride * (a0*lx+b0) + dad_x->t_offset;
	   	    ky_f = dad_y->t_stride * (a1*lx+b1) + dad_y->t_offset;

			kx = kx - dad_x->t_lb + 1;
			ky_f = ky_f - dad_y->t_lb + 1;

			ky = int_devide(kx*ty, tx);
			*amount = ky_f - ky;

			break;

		case 2 : 		// CYCLIC Distribution
		    kx = dad_x->t_stride * (a0*lx+b0) + dad_x->t_offset;
	   	    ky_f = dad_y->t_stride * (a1*lx+b1) + dad_y->t_offset;
			int px = (kx - dad_x->t_lb) % p;
			int py = (ky_f - dad_y->t_lb) % p;
			if (px == py) *amount = 0;
			else if (px > py && (ky_f/p) > (kx/p))  
						*amount = py + (p - px);
				else		*amount = py - px;

			break;
  }

}

int int_devide(int s1,int s2)
{
	int s = s1 / s2;
	int q = s1 % s2;
	//	printf("%d/%d : %d,%d\n",s1,s2,s,q);
	if (q>0) s++;
	else if (q<0) s--;
	return s;
}

// 0: NONE 1: SHIFT 2: REMAP
int detect_indepdo_communication(detect_DIM* dad_x, detect_DIM* dad_y,
                              int* l, int* u, int* stride, int *a0, int *b0,
                              int* a1, int* b1, int *amount) {
  
  int numProcs = dad_x->numProcs;
  int srcNumprocs = dad_y->numProcs;
  *amount = 0;
  

//   printf("Dad_x: code=%d p=%d lb=%d ub=%d stride=%d offset=%d\n",dad_x->distCode,dad_x->numProcs,dad_x->t_lb,dad_x->t_ub,dad_x->t_stride,dad_x->t_offset);

//   printf("Dad_y: code=%d p=%d lb=%d ub=%d stride=%d offset=%d\n",dad_y->distCode,dad_y->numProcs,dad_y->t_lb,dad_y->t_ub,dad_y->t_stride,dad_y->t_offset);

//   printf("l=%d u=%d stride=%d a0=%d b0=%d a1=%d b1=%d\n",*l,*u,*stride,*a0,*b0,*a1,*b1);
  


  if (numProcs != srcNumprocs) return 2;

  if (dad_x->distCode != dad_y->distCode) return 2;

  int tx = int_devide(dad_x->t_ub - dad_x->t_lb + 1, numProcs);
  int ty = int_devide(dad_y->t_ub - dad_y->t_lb + 1, numProcs);

  switch(dad_x->distCode) {
		case 0: 		// Collapse Distribution
				*amount = 0;
				return 0;
			break;
		case 1 : {		// BLOCK Distribution
			int t1 = dad_x->t_stride * (*a0) * ty;			// ax * a0 * ty
		        int t2 = dad_y->t_stride * (*a1) * tx;			// ay * a1 * tx
			if (t1 == t2) { 
//				if (!no_communication_detect(dad_x,
//						dad_y,*l,*u,*stride,*a0,*b0,*a1,*b1)) return 0;

				calc_shift_amount(dad_x,dad_y,numProcs,*l,*a0,*b0,*a1,*b1,
								*stride,tx,ty,amount);
				return 1;
			}
			break;
			}
		case 2 : {		// CYCLIC Distribution
			int t = (dad_y->t_stride * (*a1) * (*stride)) % numProcs;
						//  ay * a1 * s (mod p)
			if (t == (dad_x->t_stride * (*a0) * (*stride) % numProcs)) {
						//  ax * a0 * s (mod p)
				calc_shift_amount(dad_x,dad_y,numProcs,*l,*a0,*b0,*a1,*b1,
								*stride,tx,ty,amount);
				if (*amount == 0) return 0;
				else return 1;
			}
			break;
			}
	}

  return 2 ;
}

int processor_index(detect_DIM* dad_x,int tx,int sub){
	int template_index=sub*dad_x->t_stride+dad_x->t_offset;
	int offset=template_index-dad_x->t_lb+1;
	return offset/tx*tx<offset?offset/tx+1:offset/tx;
}

void processor_bound(detect_DIM* dad_x,int tx,int l,int u,int s,int *proc_lb,int *proc_ub){
	*proc_lb=processor_index(dad_x,tx,l);
	int u1=(u-l)/s*s+l;
	*proc_ub=processor_index(dad_x,tx,u1);
}

int do_section_detect(detect_DIM* dad_x, detect_DIM* dad_y,int ll,int lu,int ls,int rl,int ru,int rs,int *amount){
	int numProcs = dad_x->numProcs;
	int srcNumprocs = dad_y->numProcs;
	*amount = 0;
	if (numProcs != srcNumprocs) return 2;
	if (dad_x->distCode != dad_y->distCode) return 2;
	int tx = int_devide(dad_x->t_ub - dad_x->t_lb + 1, numProcs);
	int ty = int_devide(dad_y->t_ub - dad_y->t_lb + 1, numProcs);
	switch(dad_x->distCode){
		case 0:
			return 0;
		case 1:{ 
			int lproc_lb,lproc_ub,rproc_lb,rproc_ub;
			processor_bound(dad_x,tx,ll,lu,ls,&lproc_lb,&lproc_ub);
			if(lproc_ub>lproc_lb) return 2;
			ASSERT(lproc_ub==lproc_lb);
			processor_bound(dad_y,ty,rl,ru,rs,&rproc_lb,&rproc_ub);
			if(rproc_lb==rproc_ub&&lproc_lb==rproc_lb) return 0;
			else{ 
				if(rproc_ub>lproc_lb&&rproc_lb<lproc_lb) return 2;
				int tmp;
				if(rproc_lb>=lproc_lb){
					tmp=(ru-rl)/rs*rs+rl;
					*amount=tmp*dad_y->t_stride+dad_y->t_offset-dad_x->t_lb+1-lproc_lb*ty;
				}
				else{
					*amount=rl*dad_y->t_stride+dad_y->t_offset-dad_x->t_lb+1-((lproc_lb-1)*ty+1);
				}	
				return 1;
			}
			break;
		}	
		case 2:{
			int step=ls*dad_x->t_stride;
			if(step/numProcs*numProcs!=step&&lu>ls+ll) return 2;
			step=rs*dad_y->t_stride;
			if(step/numProcs*numProcs!=step&&ru>rs+rl) return 2;
			int llt_off=ll*dad_x->t_stride+dad_x->t_offset-dad_x->t_lb+1;
			int llproc_index=llt_off%numProcs;
			int rlt_off=rl*dad_y->t_stride+dad_y->t_offset-dad_y->t_lb+1;
			int rlproc_index=rlt_off%numProcs;
			if(llproc_index==rlproc_index) return 0;
			return 2;
			break;
		}
	}
	
}

int detect_section_communication(detect_DIM* dad_x, detect_DIM* dad_y,SgArrayRefExp *lhsarray,int lpd,SgArrayRefExp *rhsarray,int rpd,int *amount){
	SgExpression *loexp=lhsarray->subscript(lpd);
	SgExpression *roexp=rhsarray->subscript(rpd);
	SgSymbol *lhssymb=lhsarray->symbol();
	SgSymbol *rhssymb=rhsarray->symbol();
	int ll,lu,ls,rl,ru,rs;
	SgExpression *tmp;
	if(loexp->variant()==DDOT){
		if(tmp=loexp->lhs()->lhs()) ;
		else tmp=get_lbound(lhssymb,lpd);
		if(!tmp->isInteger()){
			return 3;
		}
		else ll=tmp->valueInteger();
		if(tmp=loexp->lhs()->rhs()) ;
                else tmp=get_ubound(lhssymb,lpd);
                if(!tmp->isInteger()){
			return 3;
                 }
                else lu=tmp->valueInteger();
		if(tmp=loexp->rhs()){
			if(!tmp->isInteger()){
				return 3;
                        }
			else ls=tmp->valueInteger();
		}
		else ls=1;
	}
	else {
		if(loexp->isInteger()){
			ll=lu=loexp->valueInteger();
			ls=1;
		}
		else{
			return 3;
		}
	}
	if(roexp->variant()==DDOT){
                if(tmp=roexp->lhs()->lhs()) ;
                else tmp=get_lbound(rhssymb,lpd);
		if(!tmp->isInteger()){
			return 3;
                }
                else rl=tmp->valueInteger();
                if(tmp=roexp->lhs()->rhs()) ;
                else tmp=get_ubound(rhssymb,lpd);
                if(!tmp->isInteger()){
			return 3;
                }
                else ru=tmp->valueInteger();
		if(tmp=roexp->rhs()){
                        if(!tmp->isInteger()){
				return 3;
                        }
                        else rs=tmp->valueInteger();
                }
                else rs=1;
        }
	else {
                if(roexp->isInteger()){
                        rl=ru=roexp->valueInteger();
                        rs=1;
		}
                else{
			return 3;
                }
        }
	return do_section_detect(dad_x,dad_y,ll,lu,ls,rl,ru,rs,amount);
}
